#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2020.08 #
########################################################################################
# python -m awesome_autodl.bins.show_infos --root awesome_autodl/raw_data
########################################################################################
import argparse
from pathlib import Path
from awesome_autodl.utils import load_yaml
from awesome_autodl.utils import dump_yaml
from awesome_autodl.utils import check_and_sort_by_date
from awesome_autodl.utils import filter_ele_w_value


name2file = {
    "Automated Problem Formulation": "Automated_Problem_Formulation.yaml",
    "Automated Data Engineering": "Automated_Data_Engineering.yaml",
    "Neural Architecture Search": "Neural_Architecture_Search.yaml",
    "Hyperparameter Optimization": "Hyperparameter_Optimization.yaml",
    "Automated Deployment": "Automated_Deployment.yaml",
    "Automated Maintenance": "Automated_Maintenance.yaml",
}

# add abbreviation
name2file["ADE"] = name2file["Automated Data Engineering"]
name2file["NAS"] = name2file["Neural Architecture Search"]
name2file["HPO"] = name2file["Hyperparameter Optimization"]
name2file["AD"] = name2file["Automated Deployment"]
name2file["AM"] = name2file["Automated Maintenance"]


def scale(xlist, scale):
    for x in xlist:
        print(int(x * scale))


def show_nas(root):
    abbrv = load_yaml(root / "abbrv.yaml")
    topic_path = root / "papers" / name2file["NAS"]
    assert topic_path.exists(), f"Did not find {topic_path}"
    print(f"Process NAS topic from {topic_path}")
    data = load_yaml(topic_path)
    data = check_and_sort_by_date(data)
    print(f"Find {len(data)} papers for NAS")
    # Search Space Analysis
    nasnet_like_search_space_papers = filter_ele_w_value(data, "search_space", "NASNet")
    print(
        f"NASNet-like search space: {len(nasnet_like_search_space_papers)}/{len(data)} : {len(nasnet_like_search_space_papers)*100./len(data):.3f}"
    )
    mbconv_based_search_space_papers = filter_ele_w_value(
        data, "search_space", "MBConv"
    )
    print(
        f"MBConv-based search space: {len(mbconv_based_search_space_papers)}/{len(data)} : {len(mbconv_based_search_space_papers)*100./len(data):.3f}"
    )
    size_based_search_space_papers = filter_ele_w_value(data, "search_space", "size")
    print(
        f"Size-related search space: {len(size_based_search_space_papers)}/{len(data)} : {len(size_based_search_space_papers)*100./len(data):.3f}"
    )
    ratio_keys = ["NASNet", "MBConv", "size"]
    ratios = [len(filter_ele_w_value(data, "search_space", key)) for key in ratio_keys]
    ratios = [float(ratio) / len(data) for ratio in ratios]
    ratios.append(1 - sum(ratios))
    # Search Strategy
    ratio_keys = ["Differential", "RL", "Evolution"]
    ratios = [
        len(filter_ele_w_value(data, "search_strategy", key)) for key in ratio_keys
    ]
    ratios = [float(ratio) / len(data) for ratio in ratios]
    ratios.append(1 - sum(ratios))
    # scale(ratios, 500)

    # Efficient candidate evaluation
    weight_sharing_papers = (
        filter_ele_w_value(data, "eval_boost", "weight sharing")
        + filter_ele_w_value(data, "eval_boost", "HyperNet")
        + filter_ele_w_value(data, "eval_boost", "weight i")
        + filter_ele_w_value(data, "eval_boost", "Net2Net")
    )
    print(
        f"Weight sharing: {len(weight_sharing_papers)}/{len(data)} : {len(weight_sharing_papers)*100./len(data):.3f}"
    )


def show_all(root):

    name2data = dict()
    total = 0
    for name, file in name2file.items():
        if len(name) < 10:
            topic_path = root / "papers" / file
            data = load_yaml(topic_path)
            name2data[name] = data
            total += len(data)
    for name, data in name2data.items():
        print(f"{name:5s}: {len(data)}/{total} = {len(data)*1./total:.3f}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Analysis the AutoDL papers.")
    parser.add_argument(
        "--root",
        type=str,
        required=True,
        help="The path for the data directory.",
    )
    args = parser.parse_args()
    root = Path(args.root)
    show_all(root)
    # show_nas(root)



##################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2021 #
##################################################
from awesome_autodl.utils.yaml import load_yaml
from awesome_autodl.utils.yaml import dump_yaml
from awesome_autodl.utils.check import check_paper_and_correct_format
from awesome_autodl.utils.check import check_and_sort_by_date
from awesome_autodl.utils.filter import filter_ele_w_value



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2022.01 #
########################################################################################
# python -m awesome_autodl.bins.list_email                                             #
########################################################################################
import argparse
from collections import OrderedDict, defaultdict
from awesome_autodl import autodl_topic2papers


def generate_email(author, counts):
    message = (
        f"Dear {author},"
        + "\n\n"
        + "We at the UTS' Complex Adaptive Systems Lab, part of the Data Science Institute in Sydney, Australia, have recently compiled a comprehensive conceptual review of automated deep learning (AutoDL), see initial version at https://arxiv.org/abs/2112.09245."
        + "\n\n"
        + "The intent of this review is to establish a broader snapshot of the AutoDL field beyond just neural architecture search (NAS), as well as motivate a discussion on how best to evaluate the quality of AutoDL research."
        + "\n\n"
    )
    if counts >= 3:
        message += f"In compiling this review, we have found more than {counts} of your works (referenced in the review) to be significant for informing some of the content, and we hope you find the paper of interest."
    else:
        message += "In compiling this review, we have found your work (referenced in the review) to be significant for informing some of the content, and we hope you find the paper of interest."
    message += (
        "\n\n"
        + "We thus welcome any feedback on the accuracy of our coverage - particularly with regard to your referenced contributions - and invite a broader discussion, if desired."
        + "\n\n"
        + "If you have moved on from this research area or are otherwise uninterested, we apologise for any inconvenience."
        + "\n\n"
        + "Best regards,"
        + "\n"
        + "Xuanyi Dong, David J. Kedziora, Kaska Musial and Bogdan Gabrys"
    )
    return message


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Analysis the AutoDL papers.")
    parser.add_argument(
        "--output_file",
        type=str,
        help="The path to save the final directory.",
    )
    args = parser.parse_args()

    author_email_title = []
    topic2papers = autodl_topic2papers()
    for topic, papers in topic2papers.items():
        # print(f'Collect {len(papers)} papers for "{topic}"')
        for paper in papers:
            if not paper.discussed:
                continue
            for author, email in paper.author_email.items():
                # print(f"{author:25s}, {email:15s} : {paper.title}")
                assert email is not None, "f{paper} has a None value for email"
                author_email_title.append((author, email.lower(), paper.title))
    # print(f"There are {len(author_email_title)} items in total.")

    author2email = OrderedDict()
    author2counts = defaultdict(lambda: 0)
    for author, email, title in author_email_title:
        if author in author2email:
            assert (
                author2email[author] == email
            ), f"{author} : {author2email[author]} vs {email}"
        author2email[author] = email
        author2counts[author] += 1
    # print(f"There are {len(author2email)} unique authors.")

    for author, email in author2email.items():
        message = generate_email(author, author2counts[author])
        print(f"\n\nemail:to:{email}")
        print(f"message:\n{message}")



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2021.09 #
#####################################################################################
# Automated Deep Learning: Neural Architecture Search Is Not the End, arXiv 2021.12 #
#####################################################################################
# This package is used to analyze the AutoDL-related papers. More detailed reports  #
# can be found in the above paper.                                                  #
#####################################################################################
from pathlib import Path
from collections import OrderedDict


def version():
    versions = ["v0.1"]  # 2021.09.03
    versions = ["v0.2"]  # 2021.09.04
    versions = ["v0.3"]  # 2022.01.17
    versions = ["v1.0"]  # 2022.01.20
    versions = ["v1.1"]  # 2022.01.21
    return versions[-1]


def autodl_topic2file():
    topic2file = OrderedDict()
    topic2file["Automated Problem Formulation"] = "Automated_Problem_Formulation.yaml"
    topic2file["Automated Data Engineering"] = "Automated_Data_Engineering.yaml"
    topic2file["Neural Architecture Search"] = "Neural_Architecture_Search.yaml"
    topic2file["Hyperparameter Optimization"] = "Hyperparameter_Optimization.yaml"
    topic2file["Automated Deployment"] = "Automated_Deployment.yaml"
    topic2file["Automated Maintenance"] = "Automated_Maintenance.yaml"
    return topic2file


def root():
    return Path(__file__).parent


def get_data_dir():
    return root() / "raw_data"


def get_bib_abbrv_file():
    return get_data_dir() / "abbrv.bib"


def autodl_topic2path():
    topic2file = autodl_topic2file()
    topic2path = OrderedDict()
    xdir = get_data_dir() / "papers"
    for topic, file_name in topic2file.items():
        topic2path[topic] = xdir / file_name
        if not topic2path[topic].exists():
            ValueError(f"Can not find {topic} at {topic2path[topic]}")
    return topic2path


def autodl_topic2papers():
    from awesome_autodl.utils import load_yaml, dump_yaml
    from awesome_autodl.data_cls import AutoDLpaper

    topic2path = autodl_topic2path()
    topic2papers = OrderedDict()
    for topic, xpath in topic2path.items():
        if not xpath.exists():
            ValueError(f"Can not find {topic} at {xpath}.")
        papers = []
        raw_data = load_yaml(xpath)
        assert isinstance(
            raw_data, (list, tuple)
        ), f"invalid type of raw data: {type(raw_data)}"
        for per_data in raw_data:
            papers.append(AutoDLpaper(per_data))
        topic2papers[topic] = papers
        print(f"Load {topic} completed with {len(papers)} papers.")
    return topic2papers


def get_bib_abbrv_obj():
    from awesome_autodl.data_cls import BibAbbreviations

    xfile = str(get_bib_abbrv_file())
    return BibAbbreviations(xfile)



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2020.08 #
#####################################################
from copy import deepcopy


def check_paper_and_correct_format(paper):
    assert isinstance(paper, dict), f"Expect dict than {type(paper)}"
    paper = deepcopy(paper)
    necessary_keys = (
        "title",
        "search_space",
        "search_strategy",
        "eval_boost",
        "online_date",
        "venue",
        # "application",
    )
    for key in necessary_keys:
        assert (
            key in paper
        ), f"Did not find {key} in {paper['title']} {list(paper.keys())}"
        if key != "title" and isinstance(paper[key], str):
            paper[key] = ",".join(paper[key].split(", "))
    search_strategies = (
        "RL",
        "Evolution",
        "Random",
        "Differential",
        "BayesOpt",
        "Heuristic",
        "Manual",
    )
    for search_strategy in paper["search_strategy"].split(","):
        assert (
            search_strategy in search_strategies
        ), f"This paper {paper} has a different search strategy than {search_strategies}"
    assert len(paper["online_date"]), "This paper has empty online_date"
    return paper


def check_and_sort_by_date(paper_list):
    assert isinstance(paper_list, list)
    xlist = list()
    for paper in paper_list:
        paper = check_paper_and_correct_format(paper)
        xlist.append(paper)
    return sorted(xlist, key=lambda x: x["online_date"])



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2020.08 #
#####################################################
import yaml
from pathlib import Path


def load_yaml(file_path):
    with open(str(file_path), "r") as cfile:
        data = yaml.safe_load(cfile)
    return data


def dump_yaml(data, indent=2, path=None):
    class NoAliasSafeDumper(yaml.SafeDumper):
        def ignore_aliases(self, data):
            return True

    xstr = yaml.dump(
        data,
        None,
        allow_unicode=True,
        Dumper=NoAliasSafeDumper,
        indent=indent,
        sort_keys=False,
    )
    if path is not None:
        parent_dir = Path(path).resolve().parent
        parent_dir.mkdir(parents=True, exist_ok=True)
        with open(str(path), "w") as cfile:
            cfile.write(xstr)
    return xstr



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2022.01 #
#####################################################
import re
import os


class BibAbbreviations:
    """A class to maintain the paper venue abbreviation."""

    def __init__(self, xfile):
        xfile = str(xfile)
        assert os.path.isfile(xfile)
        with open(xfile) as f:
            lines = f.readlines()
        lines = [x.strip() for x in lines]

        pattern = r'^@STRING\{([A-Z]|[a-z]|\s|_)*=(\s)*"([A-Z]|[a-z]|\s|\(|\)|\/|\,|\:|\-)*"\}$'
        self.prog = re.compile(pattern)
        self.abbrv2str = dict()
        for index, line in enumerate(lines):
            if line:
                if self.prog.match(line) is None:
                    raise ValueError(f"Incorrect line [{index}]: {line}")
                key, value = self.decode(line)
                if key in self.abbrv2str:
                    raise ValueError(f"Already defined {key}")
                self.abbrv2str[key] = value

    def __contains__(self, key):
        return key in self.abbrv2str

    def __getitem__(self, key):
        return self.abbrv2str[key]

    def __len__(self):
        return len(self.abbrv2str)

    def keys(self, sort=False):
        keys = list(self.abbrv2str.keys())
        if sort:
            keys = sorted(keys)
        return keys

    def decode(self, line):
        head = "@STRING{"
        assert len(line) > len(head)
        assert line[: len(head)] == head and line[-1] == "}"
        line = line[len(head) : -1]
        key, value = line.split("=")
        key, value = key.strip(" "), value.strip(" ")
        return key, value

    def __repr__(self):
        return f"{self.__class__.__name__}(" + f"{len(self)} abbrev pairs)"



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2022.01 #
#####################################################
from typing import Dict, Text, Any
from typing import Optional
import re


class AutoDLpaper:
    """A class to maintain the paper attribute."""

    # required fields, must include
    title: Text = None
    venue: Text = None
    venue_date: Text = None  # yyyy.mm
    online_date: Text = None  # yyyy.mm
    contacts: Optional[Dict[Text, Text]] = None  # a pair of name and email

    required_fields = ("title", "venue", "venue_date", "online_date", "contacts")

    # none-required fields
    search_space: Text = None
    search_strategy: Text = None
    candidate_evaluation: Text = None

    autodl_aspect_fields = ("search_space", "search_strategy", "candidate_evaluation")

    # non-required fileds
    links: Optional[Dict[Text, Text]] = None

    # misc
    discussed: bool = None
    misc: Text = None

    def __init__(self, data: Dict[Text, Any]):
        self.check_raw_data(data)
        self.reset_value(data)

    def reset_value(self, data):
        # set the basic value
        for field in self.required_fields[:-1]:
            if not isinstance(data[field], str):
                raise TypeError(
                    f"Expect {field} is str instead of {type(data[field])}."
                )
        self.title = data["title"]
        self.venue = data["venue"]
        date_pattern = re.compile("^([0-9]){4}.([0-9]){2}$")
        if date_pattern.match(data["venue_date"]) is None:
            raise ValueError(f"Invalid venue date ({data['venue_date']}) :: {data}")
        if date_pattern.match(data["online_date"]) is None:
            raise ValueError(f"Invalid online date ({data['online_date']}) :: {data}")
        self.venue_date = data["venue_date"]
        self.online_date = data["online_date"]

        # set the contact information
        if data["contacts"] is not None:
            assert isinstance(data["contacts"], dict)
            self.contacts = data["contacts"]
            # TODO(xuanyidong): check the email address
            # regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'

        # set the AutoDL aspect fields
        for field in self.autodl_aspect_fields:
            if data[field] is None:
                continue
            assert isinstance(data[field], str)
            setattr(self, field, data[field])

        # set the misc info
        if "discussed" in data:
            assert isinstance(
                data["discussed"], bool
            ), f"The discussed field must be a bool instead of a {type(data['discussed'])}"
            self.discussed = data["discussed"]
        if "misc" in data and data["misc"] is not None:
            assert isinstance(
                data["misc"], str
            ), f"The misc field must be a str instead of a {type(data['discussed'])}"
            self.misc = data["misc"]

    def check_raw_data(self, data):
        """Check whether the necessary field is included in `data`."""
        if not isinstance(data, dict):
            raise TypeError(f"Expect data to be a dict instead of {type(data)}")
        for field in self.required_fields:
            if field not in data:
                raise ValueError(f"Missing {field} in {data}")
        for field in self.autodl_aspect_fields:
            if field not in data:
                raise ValueError(
                    f"Missing {field} in {data}"
                    + "Please leave this field as blank if you are not sure about it."
                )
        all_fields = (
            list(self.required_fields)
            + list(self.autodl_aspect_fields)
            + ["discussed", "misc", "links"]
        )
        for key in data.keys():
            if key not in all_fields:
                raise ValueError(f"Find unexpected field: {key} in {data}")

    @property
    def author_email(self):
        if self.contacts is None:
            return dict()
        else:
            return self.contacts

    def __repr__(self):
        return f"{self.__class__.__name__}({self.title})"



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2021.09 #
#####################################################
"""The setup function for pypi."""
# The following is to make nats_bench avaliable on Python Package Index (PyPI)
#
# conda install -c conda-forge twine  # Use twine to upload nats_bench to pypi
#
# python setup.py sdist bdist_wheel
# python setup.py --help-commands
# twine check dist/*
#
# twine upload --repository-url https://test.pypi.org/legacy/ dist/*
# twine upload dist/*
# https://pypi.org/project/awesome_autodl
#
#
# or install from local: python setup.py install --force
#
import os
import glob
from pathlib import Path
from setuptools import setup, find_packages
from awesome_autodl import version

NAME = "awesome_autodl"
REQUIRES_PYTHON = ">=3.6"
DESCRIPTION = "Package for Automated Deep Learning Paper Analysis"

VERSION = version()


def read(fname="README.md"):
    with open(
        os.path.join(os.path.dirname(__file__), fname), encoding="utf-8"
    ) as cfile:
        return cfile.read()


# What packages are required for this module to be executed?
REQUIRED = ["pyyaml>=5.0.0"]

packages = find_packages(exclude=("tests", "checklist", "docs"))
print(f"packages: {packages}")


def endswith(xstring, targets):
    assert isinstance(
        targets, (list, tuple)
    ), f"invalid type of targets: {type(targets)}"
    for target in targets:
        if xstring.endswith(target):
            return True
    return False


def recursive_find_file(xdir, cur_depth=1, max_depth=1, suffixs=None):
    assert isinstance(suffixs, (list, tuple)) and len(
        suffixs
    ), f"invalid suffixs of {suffixs}"
    xdirs = []
    for xfile in Path(xdir).glob("*"):
        if xfile.is_dir() and cur_depth < max_depth:
            xdirs += recursive_find_file(xfile, cur_depth + 1, max_depth, suffixs)
        elif endswith(xfile.name, suffixs):
            xdirs.append(str(xfile))
    return xdirs


def find_yaml(xstring):
    return recursive_find_file(xstring, suffixs=(".yaml",))


def find_yaml_bib(xstring):
    return recursive_find_file(xstring, suffixs=(".yaml", ".bib"))


setup(
    name=NAME,
    version=VERSION,
    author="Xuanyi Dong",
    author_email="dongxuanyi888@gmail.com",
    description=DESCRIPTION,
    license="MIT Licence",
    keywords="NAS Dataset API DeepLearning",
    url="https://github.com/D-X-Y/Awesome-AutoDL",
    packages=packages,
    data_files=[
        (f"{NAME}/raw_data", find_yaml_bib(f"{NAME}/raw_data")),
        (f"{NAME}/raw_data/papers", find_yaml(f"{NAME}/raw_data/papers")),
    ],
    install_requires=REQUIRED,
    python_requires=REQUIRES_PYTHON,
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Database",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
    ],
)



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2020.08 #
#####################################################
from copy import deepcopy


def filter_ele_w_value(paper_list, key, value):
    assert isinstance(paper_list, list)
    xlist = list()
    for paper in paper_list:
        if value in paper[key]:
            xlist.append(deepcopy(paper))
    return xlist



#####################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2020.08 #
########################################################################################
# python -m awesome_autodl.bins.statistics --topic "ADE" --root awesome_autodl/raw_data
# python -m awesome_autodl.bins.statistics --topic "NAS" --root awesome_autodl/raw_data
# python -m awesome_autodl.bins.statistics --topic "HPO" --root awesome_autodl/raw_data
# python -m awesome_autodl.bins.statistics --topic "AD"  --root awesome_autodl/raw_data
# python -m awesome_autodl.bins.statistics --topic "AM"  --root awesome_autodl/raw_data
########################################################################################
import argparse
from pathlib import Path
from awesome_autodl.utils import load_yaml
from awesome_autodl.utils import dump_yaml
from awesome_autodl.utils import check_and_sort_by_date


name2file = {
    "Automated Problem Formulation": "Automated_Problem_Formulation.yaml",
    "Automated Data Engineering": "Automated_Data_Engineering.yaml",
    "Neural Architecture Search": "Neural_Architecture_Search.yaml",
    "Hyperparameter Optimization": "Hyperparameter_Optimization.yaml",
    "Automated Deployment": "Automated_Deployment.yaml",
    "Automated Maintenance": "Automated_Maintenance.yaml",
}

# add abbreviation
name2file["ADE"] = name2file["Automated Data Engineering"]
name2file["NAS"] = name2file["Neural Architecture Search"]
name2file["HPO"] = name2file["Hyperparameter Optimization"]
name2file["AD"] = name2file["Automated Deployment"]
name2file["AM"] = name2file["Automated Maintenance"]


def main(root, topic):
    abbrv = load_yaml(root / "abbrv.yaml")
    topic_path = root / "papers" / name2file[topic]
    assert topic_path.exists(), f"Did not find {topic_path}"
    print(f"Process {topic_path}")
    data = load_yaml(topic_path)
    data = check_and_sort_by_date(data)
    print(f"Find {len(data)} papers for {topic}")
    dump_yaml(data, path=topic_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Analysis the AutoDL papers.")
    parser.add_argument(
        "--root",
        type=str,
        required=True,
        help="The path for the data directory.",
    )
    parser.add_argument(
        "--topic",
        type=str,
        required=True,
        choices=list(name2file.keys()),
        help="Choose the AutoDL sub-topic.",
    )
    args = parser.parse_args()
    main(Path(args.root), args.topic)
